from mcap_ros2idl_support.message_definition import MessageDefinitionField
from mcap_ros2idl_support.omgidl_parser.parse import (
    Constant,
    Enum,
    Field,
    Module,
    Struct,
    Typedef,
    Union,
    UnionCase,
    parse_idl,
)
from mcap_ros2idl_support.omgidl_parser.process import (
    IDLMessageDefinition,
    IDLModuleDefinition,
    IDLStructDefinition,
    IDLUnionDefinition,
    build_map,
    parse_idl_message_definitions,
    to_idl_message_definitions,
)

__all__ = [
    "parse_idl",
    "Field",
    "Struct",
    "Module",
    "Constant",
    "Enum",
    "Typedef",
    "Union",
    "UnionCase",
    "MessageDefinitionField",
    "IDLStructDefinition",
    "IDLModuleDefinition",
    "IDLUnionDefinition",
    "IDLMessageDefinition",
    "build_map",
    "to_idl_message_definitions",
    "parse_idl_message_definitions",
]



import struct

from mcap.reader import make_reader
from mcap.records import Schema
from mcap.writer import Writer

from mcap_ros2idl_support import Ros2DecodeFactory

CDR_LE_V1_HEADER = b"\x00\x01\x00\x00"  # CDR little-endian header (version 1)


def test_decode_factory_integration(tmp_path):
    factory = Ros2DecodeFactory()

    mcap_path = tmp_path / "test.mcap"
    with open(mcap_path, "wb") as f:
        writer = Writer(f)
        writer.start()
        schema_data = b"uint32 value\n"
        schema_id = writer.register_schema("Msg", "ros2msg", schema_data)
        channel_id = writer.register_channel("/test", "cdr", schema_id)
        data = CDR_LE_V1_HEADER + struct.pack("<I", 42)
        writer.add_message(channel_id, 0, data, 0)
        writer.finish()

    with open(mcap_path, "rb") as f:
        reader = make_reader(f, decoder_factories=[factory])
        decoded = list(reader.iter_decoded_messages())
        assert len(decoded) == 1
        _, _, _, msg = decoded[0]
        assert msg == {"value": 42}


def test_decode_factory_enum_as_string(tmp_path):
    factory = Ros2DecodeFactory(enum_as_string=True)

    mcap_path = tmp_path / "test_enum.mcap"
    with open(mcap_path, "wb") as f:
        writer = Writer(f)
        writer.start()
        schema_data = (
            b"module example { enum Status { UNKNOWN, OK }; "
            b"struct Msg { example::Status status; }; };"
        )
        schema_id = writer.register_schema("example/Msg", "ros2idl", schema_data)
        channel_id = writer.register_channel("/test", "cdr", schema_id)
        import struct

        data = CDR_LE_V1_HEADER + struct.pack("<I", 1)
        writer.add_message(channel_id, 0, data, 0)
        writer.finish()

    with open(mcap_path, "rb") as f:
        reader = make_reader(f, decoder_factories=[factory])
        decoded = list(reader.iter_decoded_messages())
        assert len(decoded) == 1
        _, _, _, msg = decoded[0]
        assert msg == {"status": "OK"}


def test_decoder_caches_unsupported_schema(monkeypatch):
    factory = Ros2DecodeFactory()
    schema = Schema(
        id=1,
        name="example/Invalid",
        encoding="unknown",
        data=b"",
    )

    assert factory.decoder_for("cdr", schema) is None

    def fail(*args, **kwargs):  # pragma: no cover - fail if called
        raise AssertionError("parse_ros2idl called again")

    monkeypatch.setattr("mcap_ros2idl_support.decode_factory.parse_ros2idl", fail)

    assert factory.decoder_for("cdr", schema) is None



from mcap_ros2idl_support.rosmsg2_serialization.message_reader import (
    MessageReader,
    MessageReaderOptions,
)
from mcap_ros2idl_support.rosmsg2_serialization.message_writer import MessageWriter

__all__ = ["MessageReader", "MessageReaderOptions", "MessageWriter"]



"""Utility helpers for :class:`~cdr.encapsulation_kind.EncapsulationKind`."""

from __future__ import annotations

from dataclasses import dataclass

from mcap_ros2idl_support.cdr.encapsulation_kind import EncapsulationKind


@dataclass(frozen=True)
class EncapsulationInfo:
    """Information about the behaviour of an :class:`EncapsulationKind`.

    Attributes correspond to the flags described in the TypeScript
    implementation.  They are primarily used by the reader and writer to
    determine how a payload should be serialised or parsed.
    """

    is_cdr2: bool
    little_endian: bool
    uses_delimiter_header: bool
    uses_member_header: bool


def get_encapsulation_kind_info(kind: EncapsulationKind) -> EncapsulationInfo:
    """Return information about a given ``EncapsulationKind``.

    Parameters
    ----------
    kind:
        The :class:`EncapsulationKind` to inspect.

    Returns
    -------
    EncapsulationInfo
        A data object describing whether the kind is CDR2, little endian
        and whether it uses delimiter or member headers.
    """

    # ``Enum`` members do not support ordering comparisons directly, so we
    # compare using their underlying integer values.
    is_cdr2 = kind.value > EncapsulationKind.PL_CDR_LE.value

    little_endian = kind in {
        EncapsulationKind.CDR_LE,
        EncapsulationKind.PL_CDR_LE,
        EncapsulationKind.CDR2_LE,
        EncapsulationKind.PL_CDR2_LE,
        EncapsulationKind.DELIMITED_CDR2_LE,
        EncapsulationKind.RTPS_CDR2_LE,
        EncapsulationKind.RTPS_PL_CDR2_LE,
        EncapsulationKind.RTPS_DELIMITED_CDR2_LE,
    }

    is_delimited_cdr2 = kind in {
        EncapsulationKind.DELIMITED_CDR2_BE,
        EncapsulationKind.DELIMITED_CDR2_LE,
        EncapsulationKind.RTPS_DELIMITED_CDR2_BE,
        EncapsulationKind.RTPS_DELIMITED_CDR2_LE,
    }

    is_pl_cdr2 = kind in {
        EncapsulationKind.PL_CDR2_BE,
        EncapsulationKind.PL_CDR2_LE,
        EncapsulationKind.RTPS_PL_CDR2_BE,
        EncapsulationKind.RTPS_PL_CDR2_LE,
    }

    is_pl_cdr1 = kind in {
        EncapsulationKind.PL_CDR_BE,
        EncapsulationKind.PL_CDR_LE,
    }

    uses_delimiter_header = is_delimited_cdr2 or is_pl_cdr2
    uses_member_header = is_pl_cdr2 or is_pl_cdr1

    return EncapsulationInfo(
        is_cdr2=is_cdr2,
        little_endian=little_endian,
        uses_delimiter_header=uses_delimiter_header,
        uses_member_header=uses_member_header,
    )



"""ROS Time and Duration utilities."""

from __future__ import annotations

import math
import re
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Optional


@dataclass
class Time:
    """Represents a ROS time with second and nanosecond fields."""

    sec: int
    nsec: int


Duration = Time


def is_time(obj: object | None) -> bool:
    """Return ``True`` if *obj* looks like a :class:`Time`."""

    if isinstance(obj, Time):
        return True
    if isinstance(obj, dict):
        return set(obj.keys()) == {"sec", "nsec"}
    return False


def to_string(stamp: Time, allow_negative: bool = False) -> str:
    if not allow_negative and (stamp.sec < 0 or stamp.nsec < 0):
        raise ValueError(
            f"Invalid negative time {{ sec: {stamp.sec}, nsec: {stamp.nsec} }}"
        )
    return f"{int(stamp.sec)}.{int(stamp.nsec):09d}"


def _parse_nanoseconds(digits: str) -> int:
    digits_short = 9 - len(digits)
    return round(int(digits, 10) * 10**digits_short)


def from_string(text: str) -> Optional[Time]:
    if re.fullmatch(r"\d+\.?", text):
        sec = int(text.rstrip("."))
        return Time(sec=sec, nsec=0)
    if not re.fullmatch(r"\d+\.\d+", text):
        return None
    whole, frac = text.split(".", 1)
    sec = int(whole)
    nsec = _parse_nanoseconds(frac)
    return fix_time(Time(sec=sec, nsec=nsec))


def to_rfc3339_string(stamp: Time) -> str:
    if stamp.sec < 0 or stamp.nsec < 0:
        raise ValueError(
            f"Invalid negative time {{ sec: {stamp.sec}, nsec: {stamp.nsec} }}"
        )
    if stamp.nsec >= 1_000_000_000:
        raise ValueError(f"Invalid nanosecond value {stamp.nsec}")
    date = datetime.fromtimestamp(stamp.sec, tz=timezone.utc)
    return date.strftime("%Y-%m-%dT%H:%M:%S") + f".{stamp.nsec:09d}Z"


_RFC3339_RE = re.compile(
    r"^(\d{4,})-(\d\d)-(\d\d)[Tt](\d\d):(\d\d):(\d\d)"
    r"(?:\.(\d+))?(?:[Zz]|([+-])(\d\d):(\d\d))$"
)


def from_rfc3339_string(text: str) -> Optional[Time]:
    m = _RFC3339_RE.fullmatch(text)
    if not m:
        return None
    year, month, day, hour, minute, second, frac, sign, off_h, off_m = m.groups()
    dt = datetime(
        int(year),
        int(month),
        int(day),
        int(hour),
        int(minute),
        int(second),
        tzinfo=timezone.utc,
    )
    if sign:
        offset = timedelta(hours=int(off_h), minutes=int(off_m))
        if sign == "+":
            dt -= offset
        else:
            dt += offset
    sec = int(dt.timestamp())
    nsec = _parse_nanoseconds(frac) if frac else 0
    return fix_time(Time(sec=sec, nsec=nsec))


def to_date(stamp: Time) -> datetime:
    return datetime.fromtimestamp(stamp.sec + stamp.nsec / 1e9, tz=timezone.utc)


def from_date(date: datetime) -> Time:
    millis = int(date.timestamp() * 1000)
    sec = millis // 1000
    nsec = (millis % 1000) * 1_000_000
    return Time(sec=sec, nsec=nsec)


def percent_of(start: Time, end: Time, target: Time) -> float:
    total = subtract(end, start)
    part = subtract(target, start)
    return to_sec(part) / to_sec(total)


def interpolate(start: Time, end: Time, fraction: float) -> Time:
    duration = subtract(end, start)
    return add(start, from_sec(fraction * to_sec(duration)))


def fix_time(t: Time, allow_negative: bool = False) -> Time:
    secs_from_nanos = t.nsec // 1_000_000_000
    new_sec = t.sec + secs_from_nanos
    new_nsec = t.nsec % 1_000_000_000
    if new_nsec < 0:
        new_nsec += 1_000_000_000
        new_sec -= 1
    result = Time(sec=new_sec, nsec=new_nsec)
    if (not allow_negative and result.sec < 0) or result.nsec < 0:
        raise ValueError(f"Cannot normalize invalid time {to_string(result, True)}")
    return result


def add(a: Time, b: Time) -> Time:
    return fix_time(Time(sec=a.sec + b.sec, nsec=a.nsec + b.nsec))


def subtract(a: Time, b: Time) -> Time:
    return fix_time(Time(sec=a.sec - b.sec, nsec=a.nsec - b.nsec), allow_negative=True)


def to_nanosec(t: Time) -> int:
    return t.sec * 1_000_000_000 + t.nsec


def to_microsec(t: Time) -> float:
    return (t.sec * 1_000_000_000 + t.nsec) / 1000


def to_sec(t: Time) -> float:
    return t.sec + t.nsec * 1e-9


def from_sec(value: float) -> Time:
    sec = math.trunc(value)
    nsec = round((value - sec) * 1_000_000_000)
    sec += math.trunc(nsec / 1_000_000_000)
    nsec %= 1_000_000_000
    return Time(sec=sec, nsec=nsec)


def from_nanosec(nsec: int) -> Time:
    return Time(sec=nsec // 1_000_000_000, nsec=nsec % 1_000_000_000)


def to_millis(t: Time, round_up: bool = True) -> int:
    seconds_millis = t.sec * 1000
    nsec_millis = t.nsec / 1_000_000
    if round_up:
        return seconds_millis + math.ceil(nsec_millis)
    return seconds_millis + math.floor(nsec_millis)


def from_millis(value: int | float) -> Time:
    sec = math.trunc(value / 1000)
    nsec = round((value - sec * 1000) * 1_000_000)
    sec += math.trunc(nsec / 1_000_000_000)
    nsec %= 1_000_000_000
    return Time(sec=sec, nsec=nsec)


def from_micros(value: int | float) -> Time:
    sec = math.trunc(value / 1_000_000)
    nsec = round((value - sec * 1_000_000) * 1000)
    sec += math.trunc(nsec / 1_000_000_000)
    nsec %= 1_000_000_000
    return Time(sec=sec, nsec=nsec)


def clamp_time(time: Time, start: Time, end: Time) -> Time:
    if compare(start, time) > 0:
        return Time(sec=start.sec, nsec=start.nsec)
    if compare(end, time) < 0:
        return Time(sec=end.sec, nsec=end.nsec)
    return Time(sec=time.sec, nsec=time.nsec)


def is_time_in_range_inclusive(time: Time, start: Time, end: Time) -> bool:
    if compare(start, time) > 0 or compare(end, time) < 0:
        return False
    return True


def compare(left: Time, right: Time) -> int:
    sec_diff = left.sec - right.sec
    return sec_diff if sec_diff != 0 else left.nsec - right.nsec


def is_less_than(left: Time, right: Time) -> bool:
    return compare(left, right) < 0


def is_greater_than(left: Time, right: Time) -> bool:
    return compare(left, right) > 0


def are_equal(left: Time, right: Time) -> bool:
    return left.sec == right.sec and left.nsec == right.nsec


__all__ = [
    "Time",
    "Duration",
    "is_time",
    "to_string",
    "from_string",
    "to_rfc3339_string",
    "from_rfc3339_string",
    "to_date",
    "from_date",
    "percent_of",
    "interpolate",
    "fix_time",
    "add",
    "subtract",
    "to_nanosec",
    "to_microsec",
    "to_sec",
    "from_sec",
    "from_nanosec",
    "to_millis",
    "from_millis",
    "from_micros",
    "clamp_time",
    "is_time_in_range_inclusive",
    "compare",
    "is_less_than",
    "is_greater_than",
    "are_equal",
]



"""Python port of the ``cdr`` TypeScript library.

At this stage the port only includes a subset of the original project –
primarily enumerations and small utility helpers.  The public API mirrors
``src/index.ts`` of the original repository by re-exporting the
implemented components.
"""

from __future__ import annotations

from mcap_ros2idl_support.cdr.encapsulation_kind import EncapsulationKind
from mcap_ros2idl_support.cdr.get_encapsulation_kind_info import (
    EncapsulationInfo,
    get_encapsulation_kind_info,
)
from mcap_ros2idl_support.cdr.is_big_endian import is_big_endian
from mcap_ros2idl_support.cdr.length_codes import (
    LengthCode,
    LengthCodeError,
    get_length_code_for_object_size,
    length_code_to_object_sizes,
)
from mcap_ros2idl_support.cdr.reader import CdrReader
from mcap_ros2idl_support.cdr.reserved_pids import EXTENDED_PID, SENTINEL_PID
from mcap_ros2idl_support.cdr.size_calculator import CdrSizeCalculator
from mcap_ros2idl_support.cdr.writer import CdrWriter

__all__ = [
    "EncapsulationKind",
    "EncapsulationInfo",
    "get_encapsulation_kind_info",
    "is_big_endian",
    "LengthCode",
    "LengthCodeError",
    "get_length_code_for_object_size",
    "length_code_to_object_sizes",
    "CdrReader",
    "CdrWriter",
    "CdrSizeCalculator",
    "EXTENDED_PID",
    "SENTINEL_PID",
]



"""Enumerations for CDR encapsulation kinds.

This mirrors the constants defined in the TypeScript library's
``EncapsulationKind`` enum.  The values originate from the
DDS-XTypes specification and are used to describe how data is
serialized within a CDR stream.
"""

from __future__ import annotations

from enum import Enum


class EncapsulationKind(Enum):
    """Enumeration of the various CDR encapsulation kinds.

    The member values match those used by the TypeScript implementation
    and the DDS-XTypes specification.  ``Enum`` is used instead of
    :class:`~enum.IntEnum` to provide a closer analogue to TypeScript's
    ``enum`` semantics; code that requires the numerical value should use
    the :attr:`value` attribute of each member.
    """

    # Plain CDR encodings
    CDR_BE = 0x00  # Plain CDR, big-endian
    CDR_LE = 0x01  # Plain CDR, little-endian

    # Parameter list CDR encodings (XCDR1)
    PL_CDR_BE = 0x02  # Parameter List CDR, big-endian
    PL_CDR_LE = 0x03  # Parameter List CDR, little-endian

    # XCDR2 encodings
    CDR2_BE = 0x10  # Plain CDR2, big-endian
    CDR2_LE = 0x11  # Plain CDR2, little-endian
    PL_CDR2_BE = 0x12  # Parameter List CDR2, big-endian
    PL_CDR2_LE = 0x13  # Parameter List CDR2, little-endian
    DELIMITED_CDR2_BE = 0x14  # Delimited CDR2, big-endian
    DELIMITED_CDR2_LE = 0x15  # Delimited CDR2, little-endian

    # RTPS specific IDs for XCDR2
    RTPS_CDR2_BE = 0x06  # Plain CDR2, big-endian
    RTPS_CDR2_LE = 0x07  # Plain CDR2, little-endian
    RTPS_DELIMITED_CDR2_BE = 0x08  # Delimited CDR2, big-endian
    RTPS_DELIMITED_CDR2_LE = 0x09  # Delimited CDR2, little-endian
    RTPS_PL_CDR2_BE = 0x0A  # Parameter List CDR2, big-endian
    RTPS_PL_CDR2_LE = 0x0B  # Parameter List CDR2, little-endian
